﻿using System;
using System.IO;
using System.Collections.Generic;
using Clarion;
using Clarion.Framework;
using Clarion.Framework.Core;
using Clarion.Framework.Extensions;
using Clarion.Framework.Templates;

namespace Blackjack
{
    class Blackjack
    {
        static string[] playerCards = new string[11];
        static string hitOrStay = "";
        static int total = 0, count = 1, dealerTotal = 0;
        static Random cardRandomizer = new Random();
        static int win = 0, loose = 0;
        static int Player = 0;
        //Common Variables
        //static DimensionValuePair hitOrDone;
        static List<DimensionValuePair<string, int>> dvs;
        static ExternalActionChunk hit;
        static ExternalActionChunk done;
        static ExternalActionChunk chosen;
        static StreamWriter sw = new System.IO.StreamWriter("statistics.txt");
        //Player Beginner
        //simple dummy agent
        static Agent PlayerBeginner;
        static SimplifiedQBPNetwork net;
        //Player Intermediate
        //fixed rules agent
        static Agent PlayerIntermediate;
        //Player Advanced
        //meta cognitive agent
        static GoalChunk gcHit;
        static GoalChunk gcDone;
        static Agent PlayerAdvanced;
        static SimplifiedQBPNetwork net2;

        private static double FixedRuleDelegateToAvoidExplode(ActivationCollection currentInput, Rule target)
        {
            return (total >= 16) ? 1.0 : 0.0;
        }
        private static double FixedRuleDelegateToDoHit(ActivationCollection currentInput, Rule target)
        {
            return (total < 16) ? 1.0 : 0.0;
        }
        
        public static void AddSensoryCards(SensoryInformation si, List<DimensionValuePair<string, int>> dvs)
        {
            for (int i = 1; i < 21; i++)
            {
                 DimensionValuePair dv = dvs[i];
                 if (i == total)
                 {
                         si.Add(dv, 1.0);
                 }
                 else
                 {
                         si.Add(dv, 0.0);
                 }
            }
        }


        public static double DeficitChange(ActivationCollection si, Drive target)
        {
            var cg = ((SensoryInformation)si).AffiliatedAgent.CurrentGoal;
            if (cg != null)
            {
                if (total > 16)
                    return 0.0; //target.Parameters.DEFICIT_CHANGE_RATE = 0.5;
                else
                    return 1.0; //target.Parameters.DEFICIT_CHANGE_RATE = 2.0;
            }

            return target.Deficit * target.Parameters.DEFICIT_CHANGE_RATE;
        }

        static void Main(string[] args)
        {
            Console.Title = "21 utilizando Clarion!";
            //sw = File.CreateText("statistics.csv");
            //sw = new System.IO.StreamWriter("statistics.csv");

            Console.WriteLine("Qual Jogador? 0, 1 ou 2?\n");
            string PlayerInput = Console.ReadLine().ToLower();
            string PlayerString = "Iniciante";

            if (PlayerInput == "0") 
            {
                Player = 0; 
                PlayerString = "Iniciante";
            }

            if (PlayerInput == "1") 
            {
                Player = 1; 
                PlayerString = "Intermediario";
            }
            if (PlayerInput == "2") 
            { 
                Player = 2; 
                PlayerString = "Avançado"; 
            }

            sw.WriteLine("Jogando com Jogador " + PlayerString + "\n"); sw.Flush();
            sw.WriteLine("total;dealerTotal;resultado\n"); sw.Flush();
            
            //hitOrDone = World.NewDimensionValuePair("Play", "hitOrDone");
            dvs = new List<DimensionValuePair<string, int>>();
            for (int i = 1; i < 21; i++)
            {
                dvs.Add(World.NewDimensionValuePair("Points", i));
            }
            
            hit = World.NewExternalActionChunk("hit");
            done = World.NewExternalActionChunk("done");

            //Initialize the Beginner Agent
            PlayerBeginner = World.NewAgent("PlayerBeginer");
            net = AgentInitializer.InitializeImplicitDecisionNetwork(PlayerBeginner, SimplifiedQBPNetwork.Factory);
            //net.Input.Add(hitOrDone);
            net.Input.AddRange(dvs);

            net.Output.Add(hit);
            net.Output.Add(done);
            PlayerBeginner.Commit(net);
            net.Parameters.LEARNING_RATE = 1;
            PlayerBeginner.ACS.Parameters.PERFORM_RER_REFINEMENT = false;

            //Initialize the Intermediate Agent
            PlayerIntermediate = World.NewAgent("PlayerIntermediate");

            SupportCalculator avoidExplode = FixedRuleDelegateToAvoidExplode;
            FixedRule ruleAvoidExplode = AgentInitializer.InitializeActionRule(PlayerIntermediate, FixedRule.Factory, done, avoidExplode);
            PlayerIntermediate.Commit(ruleAvoidExplode);

            SupportCalculator doHit = FixedRuleDelegateToDoHit;
            FixedRule ruleDoHit = AgentInitializer.InitializeActionRule(PlayerIntermediate, FixedRule.Factory, hit, doHit);
            PlayerIntermediate.Commit(ruleDoHit);

            // Disable Rule Refinement
            PlayerIntermediate.ACS.Parameters.PERFORM_RER_REFINEMENT = false;

            // The selectio type will be probabilistic
            PlayerIntermediate.ACS.Parameters.LEVEL_SELECTION_METHOD = ActionCenteredSubsystem.LevelSelectionMethods.STOCHASTIC;

            // The selection of the action will be fixed (not variable) i.e. only the statement defined above.
            PlayerIntermediate.ACS.Parameters.LEVEL_SELECTION_OPTION = ActionCenteredSubsystem.LevelSelectionOptions.FIXED;


            // Define Probabilistic values
            PlayerIntermediate.ACS.Parameters.FIXED_FR_LEVEL_SELECTION_MEASURE = 1;
            PlayerIntermediate.ACS.Parameters.FIXED_IRL_LEVEL_SELECTION_MEASURE = 0;
            PlayerIntermediate.ACS.Parameters.FIXED_BL_LEVEL_SELECTION_MEASURE = 0;
            PlayerIntermediate.ACS.Parameters.FIXED_RER_LEVEL_SELECTION_MEASURE = 0;
            //Initialize the Advanced Agent
            gcHit = World.NewGoalChunk("Hit");
            gcDone = World.NewGoalChunk("Done");

            //Initialize the Agent
            PlayerAdvanced = World.NewAgent("PlayerAdvanced");

            SimplifiedQBPNetwork net2 = AgentInitializer.InitializeImplicitDecisionNetwork(PlayerAdvanced, SimplifiedQBPNetwork.Factory);

            net2.Input.Add(gcHit);
            net2.Input.Add(gcDone);

            //net2.Input.Add(hitOrDone);
            for (int i = 1; i < 21; i++)
            {
                dvs.Add(World.NewDimensionValuePair("Points", i));
            }
            net2.Input.AddRange(dvs);

            net2.Output.Add(hit);
            net2.Output.Add(done);

            net2.Parameters.LEARNING_RATE = 1;

            PlayerAdvanced.Commit(net2);

            RefineableActionRule.GlobalParameters.SPECIALIZATION_THRESHOLD_1 = -.6;
            RefineableActionRule.GlobalParameters.GENERALIZATION_THRESHOLD_1 = -.1;
            RefineableActionRule.GlobalParameters.INFORMATION_GAIN_OPTION = RefineableActionRule.IGOptions.PERFECT;

            AutonomyDrive aut = AgentInitializer.InitializeDrive(PlayerAdvanced, AutonomyDrive.Factory, .5, (DeficitChangeProcessor)DeficitChange);

            DriveEquation autd =
                AgentInitializer.InitializeDriveComponent(aut, DriveEquation.Factory);

            aut.Commit(autd);

            PlayerAdvanced.Commit(aut);

            GoalSelectionModule gsm =
                AgentInitializer.InitializeMetaCognitiveModule(PlayerAdvanced, GoalSelectionModule.Factory);

            GoalSelectionEquation gse =
                AgentInitializer.InitializeMetaCognitiveDecisionNetwork(gsm, GoalSelectionEquation.Factory);

            gse.Input.Add(aut.GetDriveStrength());

            GoalStructureUpdateActionChunk su = World.NewGoalStructureUpdateActionChunk();
            GoalStructureUpdateActionChunk bu = World.NewGoalStructureUpdateActionChunk();

            su.Add(GoalStructure.RecognizedActions.SET_RESET, gcHit);
            bu.Add(GoalStructure.RecognizedActions.SET_RESET, gcDone);

            gse.Output.Add(su);
            gse.Output.Add(bu);

            gse.SetRelevance(bu, aut, 1);

            gsm.Commit(gse);

            PlayerAdvanced.Commit(gsm);

            PlayerAdvanced.MS.Parameters.CURRENT_GOAL_ACTIVATION_OPTION =
                MotivationalSubsystem.CurrentGoalActivationOptions.FULL;

            Start();

            sw.Close();
            PlayerBeginner.Die();
            PlayerIntermediate.Die();
            PlayerAdvanced.Die();
            
        }

        static void Start()
        {
            dealerTotal = cardRandomizer.Next(15, 22);
            playerCards[0] = Deal();
            playerCards[1] = Deal();
            SensoryInformation si;
            //List<DimensionValuePair<string, int>> dvs = new List<DimensionValuePair<string, int>>();

            do
            {
                Console.WriteLine("Jogador recebeu " + playerCards[0] + " e " + playerCards[1] + ". \nO total é " + total + ".\nContinuar (s/n)?");
                //hitOrStay = Console.ReadLine().ToLower();
        
                if (Player == 0)
                {
                    si = World.NewSensoryInformation(PlayerBeginner);
                    //si.Add(hitOrDone, PlayerBeginner.Parameters.MAX_ACTIVATION);
                    AddSensoryCards(si, dvs);

                    PlayerBeginner.Perceive(si);
                    chosen = PlayerBeginner.GetChosenExternalAction(si);
                }
                else if (Player == 1)
                {
                    si = World.NewSensoryInformation(PlayerIntermediate);
                    //si.Add(hitOrDone, PlayerBeginner.Parameters.MAX_ACTIVATION);
                    AddSensoryCards(si, dvs);

                    PlayerIntermediate.Perceive(si);
                    chosen = PlayerIntermediate.GetChosenExternalAction(si);
                }
                else //if (Player == 2)
                {
                    si = World.NewSensoryInformation(PlayerAdvanced);
                    //si.Add(hitOrDone, PlayerAdvanced.Parameters.MAX_ACTIVATION);
                    AddSensoryCards(si, dvs);
                    //si[typeof(AutonomyDrive), AutonomyDrive.MetaInfoReservations.STIMULUS] = 1;

                    PlayerAdvanced.Perceive(si);
                    chosen = PlayerAdvanced.GetChosenExternalAction(si);
                }

                //Deliver appropriate feedback to the agent
                if (chosen == hit)
                {
                    Console.WriteLine("Continuar\n");
                    hitOrStay = "s";
                }
                else //done
                {
                    Console.WriteLine("Parar\n");
                    hitOrStay = "n";
                }

                //hitOrStay = Console.ReadLine().ToLower();
            } while (!hitOrStay.Equals("s") && !hitOrStay.Equals("n"));
            Game(si);
        }

        static void Game(SensoryInformation si)
        {
            if (hitOrStay.Equals("s"))
            {
                Hit(si);
            }
            else if (hitOrStay.Equals("n"))
            {
                if (total > dealerTotal && total <= 21)
                {
                    Console.WriteLine("\nJogador venceu! Total da banca é " + dealerTotal + ".\nJogar novamente? s/n");
                    //Give positive feedback.
                    if (Player == 0)
                    {
                        PlayerBeginner.ReceiveFeedback(si, 1.0);
                    }
                    if (Player == 1)
                    {
                        //PlayerIntermediate.ReceiveFeedback(si, 1.0);
                    }
                    if (Player == 2)
                    {
                         PlayerAdvanced.ReceiveFeedback(si, 1.0);
                    }

                    win++;
                    sw.WriteLine(total + ";" + dealerTotal + ";" + "venceu\n"); sw.Flush();
                    PlayAgain();
                }
                else if (total < dealerTotal)
                {
                    Console.WriteLine("\nJogador Perdeu! Total da banca é " + dealerTotal + ".\nJogar novamente? s/n");
                    //Give positive feedback.
                    if (Player == 0)
                    {
                        PlayerBeginner.ReceiveFeedback(si, 0.0);
                    }
                    if (Player == 1)
                    {
                        //PlayerIntermediate.ReceiveFeedback(si, 0.0);
                    }
                    if (Player == 2)
                    {
                        PlayerAdvanced.ReceiveFeedback(si, 0.0);
                    } loose++;
                    sw.WriteLine(total + ";" + dealerTotal + ";" + "perdeu\n"); sw.Flush();
                    PlayAgain();
                }
                else
                {
                    Console.WriteLine("\nEmpate. Jogador Perdeu! Total da banca é " + dealerTotal + ".\nJogar novamente? s/n");
                    //Give positive feedback.
                    if (Player == 0)
                    {
                        PlayerBeginner.ReceiveFeedback(si, 0.5);
                    }
                    if (Player == 1)
                    {
                        //PlayerIntermediate.ReceiveFeedback(si, 0.5);
                    }
                    if (Player == 2)
                    {
                        PlayerAdvanced.ReceiveFeedback(si, 0.5);
                    }
                    loose++;
                    sw.WriteLine(total + ";" + dealerTotal + ";" + "Empate\n"); sw.Flush();
                    PlayAgain();
                }

            }
            Console.ReadLine();
        }

        static string Deal()
        {
            string Card = "";
            int cards = cardRandomizer.Next(1, 14);
            switch (cards)
            {
                case 1: Card = "Dois"; total += 2;
                    break;
                case 2: Card = "Tres"; total += 3;
                    break;
                case 3: Card = "Quatro"; total += 4;
                    break;
                case 4: Card = "Cinco"; total += 5;
                    break;
                case 5: Card = "Seis"; total += 6;
                    break;
                case 6: Card = "Sete"; total += 7;
                    break;
                case 7: Card = "Oito"; total += 8;
                    break;
                case 8: Card = "Nove"; total += 9;
                    break;
                case 9: Card = "Dez"; total += 10;
                    break;
                case 10: Card = "Valet"; total += 10;
                    break;
                case 11: Card = "Dama"; total += 10;
                    break;
                case 12: Card = "Reis"; total += 10;
                    break;
                case 13: Card = "As"; total += 11;
                    break;
                default: Card = "2"; total += 2;
                    break;
            }
            return Card;
        }

        static void Hit(SensoryInformation si)
        {
            count += 1;
            playerCards[count] = Deal();
            Console.WriteLine("\nJogador recebeu " + playerCards[count] + ".\nO total é " + total + ".");
            if (total.Equals(21))
            {
                Console.WriteLine("\nJogador fez 21! Total do jogador é " + total + ".\nO total da banca é " + dealerTotal + ".\nJogar novamente (s/n)?");
                //Give positive feedback.
                if (Player == 0)
                {
                    PlayerBeginner.ReceiveFeedback(si, 1.0);
                }
                if (Player == 1)
                {
                   // PlayerIntermediate.ReceiveFeedback(si, 1.0);
                }
                if (Player == 2)
                {
                    PlayerAdvanced.ReceiveFeedback(si, 1.0);
                }
                win++;
                sw.WriteLine(total + ";" + dealerTotal + ";" + "blackjack\n"); sw.Flush();
                PlayAgain();
            }
            else if (total > 21)
            {
                Console.WriteLine("\nJogador estourou! Total do jogador é " + total + ".\nO total da banca é " + dealerTotal + ".\nJogar novamente (s/n)");
                //Give positive feedback.
                if (Player == 0)
                {
                    PlayerBeginner.ReceiveFeedback(si, 0.0);
                }
                if (Player == 1)
                {
                   // PlayerIntermediate.ReceiveFeedback(si, 0.0);
                }
                if (Player == 2)
                {
                    PlayerAdvanced.ReceiveFeedback(si, 0.0);
                }
                loose++;
                sw.WriteLine(total + ";" + dealerTotal + ";" + "estourou\n"); sw.Flush();
                PlayAgain();
            }
            else if (total < 21)
            {
                if (Player == 0)
                {
                    PlayerBeginner.ReceiveFeedback(si, 1.0);
                }
                if (Player == 1)
                {
                    //PlayerIntermediate.ReceiveFeedback(si, 1.0);
                }
                if (Player == 2)
                {
                    PlayerAdvanced.ReceiveFeedback(si, 1.0);
                }
                SensoryInformation siNew;

                do
                {
                    Console.WriteLine("\nDesce?");
                    if (Player == 0)
                    {
                        siNew = World.NewSensoryInformation(PlayerBeginner);
                        //siNew.Add(hitOrDone, PlayerBeginner.Parameters.MAX_ACTIVATION);
                        AddSensoryCards(siNew, dvs);

                        PlayerBeginner.Perceive(siNew);
                        chosen = PlayerBeginner.GetChosenExternalAction(siNew);
                    }
                    if (Player == 1)
                    {
                        siNew = World.NewSensoryInformation(PlayerIntermediate);
                        //siNew.Add(hitOrDone, PlayerBeginner.Parameters.MAX_ACTIVATION);
                        AddSensoryCards(siNew, dvs);

                        PlayerIntermediate.Perceive(siNew);
                        chosen = PlayerIntermediate.GetChosenExternalAction(siNew);
                    }
                    else //if (Player == 2)
                    {
                        siNew = World.NewSensoryInformation(PlayerAdvanced);
                        //siNew.Add(hitOrDone, PlayerAdvanced.Parameters.MAX_ACTIVATION);
                        AddSensoryCards(siNew, dvs);
                        //siNew[typeof(AutonomyDrive), AutonomyDrive.MetaInfoReservations.STIMULUS] = 1;

                        PlayerAdvanced.Perceive(siNew);
                        chosen = PlayerAdvanced.GetChosenExternalAction(siNew);
                    }

                    //Deliver appropriate feedback to the agent
                    if (chosen == hit)
                    {
                        Console.WriteLine("Continuar\n");
                        hitOrStay = "s";
                    }
                    else //done
                    {
                        Console.WriteLine("Parar\n");
                        hitOrStay = "n";
                    }
                    
                    //hitOrStay = Console.ReadLine().ToLower();
                } while (!hitOrStay.Equals("s") && !hitOrStay.Equals("n"));
                Game(siNew);
            }
        }

        static void PlayAgain()
        {
            string playAgain = "";
            do
            {
                //playAgain = Console.ReadLine().ToLower();
                if ((win + loose) < 1000) playAgain = "s";
                else playAgain = "n";
            } while (!playAgain.Equals("s") && !playAgain.Equals("n"));
            if (playAgain.Equals("s"))
            {
                //Console.WriteLine("\nPressione qualquer tecla para reiniciar!");
                //Console.ReadLine();
                Console.Clear();
                dealerTotal = 0;
                count = 1;
                total = 0;
                Start();
            }
            else if (playAgain.Equals("n"))
            {
                Console.WriteLine("\nPressione qualquer tecla para sair");
                Console.ReadLine();
                Environment.Exit(0);
            }

        }
    }
}

